def test_import():
    import mlflow_algorithmia

    assert mlflow_algorithmia.__version__ is not None
    assert mlflow_algorithmia.__version__ != "0.0.0"
    assert len(mlflow_algorithmia.__version__) > 0



# This file is based on conda_env source
# because we only read conda.yaml files generated by mlflow we can be more relaxed
# and we dont have to add a dependency on conda itself

import re
from collections import OrderedDict

import pkg_resources
import ruamel.yaml as yaml
from pip._vendor.packaging.utils import canonicalize_name, canonicalize_version


def yaml_safe_load(string):
    """
    Examples:
        >>> yaml_safe_load("key: value")
        {'key': 'value'}
    """
    # Version 1.2 is the same that conda supports
    return yaml.safe_load(string, version="1.2")


class Environment(object):
    def __init__(
        self,
        name=None,
        channels=None,
        dependencies=None,
    ):
        self.name = name
        self.dependencies = Dependencies(dependencies)

        if channels is None:
            channels = []
        self.channels = channels

    @classmethod
    def from_file(cls, fpath):
        with open(fpath, "r") as file:
            yaml_str = file.read()
            return cls.from_yamlstr(yaml_str)

    @classmethod
    def from_yamlstr(cls, yamlstr, **kwargs):
        """
        Load and return a ``Environment`` from a given ``yaml string``
        """
        data = yaml_safe_load(yamlstr)
        # data = validate_keys(data, kwargs)

        if kwargs is not None:
            for key, value in kwargs.items():
                data[key] = value

        return cls(**data)

    def to_dict(self):
        d = OrderedDict([("name", self.name)])
        if self.channels:
            d["channels"] = self.channels
        if self.dependencies:
            d["dependencies"] = self.dependencies
        return d

    def list_deps(self):
        """
        Returns a single list with the dependencies we need to deploy the model
        1. Python and pip
        """
        ret = []
        ignore_list = ("python", "pip")
        for dep in self.dependencies["conda"]:
            if dep.name not in ignore_list:
                ret.append(str(dep))

        for dep in self.dependencies["pip"]:
            ret.append(str(dep))

        return ret

    def __repr__(self):
        return str(self.to_dict())


class Dependencies(OrderedDict):
    def __init__(self, raw, *args, **kwargs):
        super(Dependencies, self).__init__(*args, **kwargs)
        self.raw = raw
        self.parse()

    def parse(self):
        if not self.raw:
            return

        self.update({"conda": [], "pip": []})

        for dep in self.raw:
            if isinstance(dep, str):
                # conda deps

                # 1. Convert the conda single = to pip double ==
                dep = re.sub(r"(?<!=)=(?!=)", "==", dep)

                self["conda"].append(MatchSpec.from_str(dep))
            elif isinstance(dep, dict):
                # pip dependencies
                for dep in dep["pip"]:
                    self["pip"].append(MatchSpec.from_str(dep))


class MatchSpec(object):
    def __init__(self, requirement):
        self.requirement = requirement
        self.name = requirement.name
        self.specs = sorted(requirement.specs)

    @classmethod
    def from_str(cls, string):
        requirement = pkg_resources.parse_requirements(string)
        requirement = list(requirement)[0]
        return cls(requirement=requirement)

    def __repr__(self):
        specs = []
        for op, ver in self.specs:
            specs.append(f"{op}{ver}")

        if len(specs) > 0:
            specs = ",".join(specs)
            return f"{self.name}{specs}"
        else:
            return f"{self.name}"

    def __str__(self):
        return self.__repr__()



import json
import logging
import os
import shutil
import sys
import tarfile
import urllib
from urllib.parse import urlparse

import Algorithmia
import requests
import ruamel.yaml as yaml
from Algorithmia.errors import raiseAlgoApiError
from algorithmia_api_client.rest import ApiException
from git import Git, Repo, remote
from jinja2 import Environment, FileSystemLoader
from mlflow.deployments import BaseDeploymentClient
from mlflow.exceptions import MlflowException

from mlflow_algorithmia.conda_env import Environment as CondaEnvironment


logger = logging.getLogger(__name__)
CURDIR = os.path.dirname(os.path.realpath(__file__))


class AlgorithmiaDeploymentClient(BaseDeploymentClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.repo = None
        self.algo = None
        self.run_id = None
        self.mlmodel = None
        self.settings = Settings()
        self.client = Algorithmia.client(self.settings["api_key"])

    def create_deployment(self, name, model_uri, flavor=None, config=None):
        """
        Creates a deployment of the MLflow model in Algorithmia
        1. Creates and uploads a model bundle
        2. Creates the source code
        """
        self.settings.update(config)

        if config and config.get("raiseError") == "True":
            raise RuntimeError("Error requested")

        self.create_algorithm(name)
        self.update_deployment(
            name=name, model_uri=model_uri, flavor=flavor, config=config
        )
        return {"name": name, "flavor": "Algorithmia"}

    def update_deployment(self, name, model_uri=None, flavor=None, config=None):
        """
        Updates a deployment in Algorithmia
        1. Creates and uploads a new model bundle
        2. Updates the source code
        """
        self.read_model_metadata(model_uri)

        os.makedirs(self.settings["tmp_dir"], exist_ok=True)
        tar_file = self.create_bundle(model_uri)
        algo_tar_file = self.upload_bundle(name, tar_file)

        repo_path = self.repo_clone_or_pull(name)
        config = {
            "mlflow_bundle_file": algo_tar_file,
            "dependencies": self.get_requirements(model_uri),
        }
        self.update_source(name, repo_path, **config)

        self.repo_commit_and_push()

        version = self.get_builds(name)[0]["commit_sha"]
        logger.info("New model version ready: %s", version)
        return {"name": name, "flavor": "Algorithmia"}

    def list_deployments(self):
        return "To see Algorithmia deployments go to the Algorithmia homepage"

    def delete_deployment(self, name):
        """
        Deletes a deployment in algorithmia and removes local
        temp directory
        """
        if os.path.exists(self.settings["tmp_dir"]):
            shutil.rmtree(self.settings["tmp_dir"])
        self.delete_algorithm(name)

    def get_deployment(self, name):
        username = self.settings["username"]
        algo_namespace = f"{username}/{name}"
        algo = self.client.algo(algo_namespace)

        return {
            "name": algo.algoname,
            "username": algo.username,
            "url": algo.url,
        }

    def predict(self, deployment_name, df):
        username = self.settings["username"]
        algo_namespace = f"{username}/{deployment_name}"
        algo = self.client.algo(algo_namespace)

        query = df.to_json(orient="split")
        return algo.pipe(query).result

    # Util functions

    def create_algorithm(self, name):
        """
        Create an algorithm in Algorithmia
        """
        logger.info("Creating Algorithm %s", name)

        details = {
            "label": name,
            "summary": self.settings["summary"],
            "tagline": self.settings["tagline"],
        }
        settings = {
            "package_set": self.settings["package_set"],
            "source_visibility": self.settings["source_visibility"],
            "license": self.settings["license"],
            "network_access": self.settings["network_access"],
            "pipeline_enabled": self.settings["pipeline_enabled"],
        }

        self.algo_(name).create(details=details, settings=settings)
        logger.info("Algorithm %s created", name)

    def algo_(self, name):
        """
        Internal representation for the algorithmia algorithm
        """
        username = self.settings["username"]
        algo_namespace = f"{username}/{name}"
        return self.client.algo(algo_namespace)

    def get_builds(self, name):
        api = self.settings["api_endpoint"]
        username = self.settings["username"]
        url = f"{api}/v1/algorithms/{username}/{name}/builds"

        api_key = self.settings["api_key"]
        headers = {"Authorization": f"Simple {api_key}"}
        r = requests.get(url, headers=headers)
        return r.json()["results"]

    def delete_algorithm(self, name):
        """Deletes an algorithm in Algorithmia"""
        logger.info("Deleting %s deployment in Algorithmia", name)
        try:
            api_response = self.client.manageApi.delete_algorithm(
                self.settings["username"], name
            )
            return api_response
        except ApiException as ex:
            error_message = json.loads(ex.body)
            raiseAlgoApiError(error_message)
        logger.info("Algorithm %s deleted", name)

    def read_model_metadata(self, model_uri):
        """
        Read MLmodel file
        """
        mlmodel_path = os.path.join(model_uri, "MLmodel")
        with open(mlmodel_path, "r") as file:
            mlmodel = yaml.load(file, Loader=yaml.Loader)

        self.mlmodel = mlmodel
        self.run_id = self.mlmodel["run_id"]

    def create_bundle(self, model_uri):
        """
        Creates a .tar.gz bundle from the MLflow model
        """
        logger.info("Creating Mlflow bundle")
        tar_fname = f"model-{self.run_id}.tar.gz"
        tar_fpath = os.path.join(self.settings["tmp_dir"], tar_fname)
        with tarfile.open(tar_fpath, "w:gz") as tar:
            source_dir = model_uri
            tar.add(source_dir, arcname=tar_fname[: -len(".tar.gz")])

        return tar_fpath

    def upload_bundle(self, name, tar_fpath):
        """
        Upload the MLflow bundle to algorithmia
        """
        logger.info("Uploading Mlflow bundle")
        username = self.settings["username"]
        algo_data_dir = f"data://{username}/{name}"

        if not self.client.dir(algo_data_dir).exists():
            self.client.dir(algo_data_dir).create()

        tar_fname = os.path.basename(tar_fpath)
        algo_file = os.path.join(algo_data_dir, tar_fname)
        self.client.file(algo_file).putFile(tar_fpath)
        logger.info("MLflow bundle uploaded to: %s", algo_file)
        return algo_file

    def repo_clone_or_pull(self, name):
        """
        Clones a repository from Algorithmia to the temp directory
        If repo it's already cloned then pulls any changes
        """
        target_dir = self.settings["tmp_dir"]
        logger.info("Cloning algorithm source to: %s", target_dir)
        os.makedirs(target_dir, exist_ok=True)
        repo_path = os.path.join(target_dir, name)

        if not os.path.exists(repo_path):
            username = self.settings["username"]
            encoded_api_key = self.settings["encoded_api_key"]
            git_endpoint = self.settings["git_endpoint"]
            username = self.settings["username"]
            algo_name = name
            algo_repo = f"https://{username}:{encoded_api_key}@{git_endpoint}/git/{username}/{algo_name}.git"
            self.repo = Repo.clone_from(algo_repo, repo_path)
            # self.repo = Repo.clone_from(algo_repo, repo_path, progress=Progress())
        else:
            self.repo = Repo(repo_path)
            origin = self.repo.remote(name="origin")
            origin.pull()

        return repo_path

    def repo_commit_and_push(self):
        logger.info("Updating algorithm source and building model")
        commit_msg = f"Update - MLflow run_id: {self.run_id}"
        self.repo.git.add(".")
        self.repo.index.commit(commit_msg)
        origin = self.repo.remote(name="origin")
        origin.push()
        logger.info("Algorithm repo updated: %s", commit_msg)

    def update_source(self, name, repo_path, **kwargs):
        _ = os.path.join(repo_path, ".gitignore")
        self.render_file("gitignore_repo", _, **kwargs)

        _ = os.path.join(repo_path, "requirements.txt")
        self.render_file("requirements.txt", _, **kwargs)

        _ = os.path.join(repo_path, "src", f"{name}.py")
        self.render_file("entrypoint.py", _, **kwargs)

        _ = os.path.join(repo_path, "src", "algorithmia_utils.py")
        self.render_file("algorithmia_utils.py", _, **kwargs)

        _ = os.path.join(repo_path, "src", "mlflow_wrapper.py")
        self.render_file("mlflow_wrapper.py", _, **kwargs)

        _ = os.path.join(repo_path, "models")
        os.makedirs(_, exist_ok=True)

        _ = os.path.join(repo_path, "models", ".gitignore")
        self.render_file("gitignore_all", _, **kwargs)

    def render_file(self, inp, out, **kwargs):
        file_loader = FileSystemLoader(os.path.join(CURDIR, "templates"))
        env = Environment(loader=file_loader)
        template = env.get_template(inp)
        output = template.render(**kwargs)

        with open(out, "w") as file:
            file.write(output)

    def get_requirements(self, model_uri):
        """
        Return a list of requirements based on the MLflow conda.yaml
        """
        conda_yaml_path = os.path.join(model_uri, "conda.yaml")
        environemnt = CondaEnvironment.from_file(conda_yaml_path)
        return environemnt.list_deps()


class Settings(dict):
    def __init__(self):
        super().__init__()
        self["api_key"] = os.environ.get("ALGORITHMIA_API_KEY", None)
        self["username"] = os.environ.get("ALGORITHMIA_USERNAME", None)

        if self["api_key"] is None:
            raise MlflowException(
                "Environment variable ALGORITHMIA_API_KEY is not set."
            )

        if self["username"] is None:
            raise MlflowException(
                "Environment variable ALGORITHMIA_USERNAME is not set."
            )

        default_api = "https://api.algorithmia.com"
        self["api_endpoint"] = os.environ.get("ALGORITHMIA_API", default_api)
        # Encoding the API key, so we can use it in the git URL
        self["encoded_api_key"] = urllib.parse.quote_plus(self["api_key"])
        self["tagline"] = os.environ.get("ALGORITHM_TAGLINE", "Mlflow deployment")
        self["summary"] = os.environ.get("ALGORITHM_SUMMARY", "Mlflow deployment")

        # Make the git_endpoint from the Algorithmia API
        url = urlparse(self["api_endpoint"]).netloc
        url = url[4:] if url.startswith("www.") else url
        url = url[4:] if url.startswith("api.") else url
        default_git_endpoint = f"git.{url}"
        self["git_endpoint"] = os.environ.get(
            "ALGORITHMIA_GIT_ENDPOINT", default_git_endpoint
        )

        default_tmp_dir = "./algorithmia_tmp/"
        self["tmp_dir"] = os.environ.get("MLFLOW_ALGO_TMP_DIR", default_tmp_dir)

        self["package_set"] = os.environ.get("ALGO_PACKAGE_SET", "python37")
        self["source_visibility"] = os.environ.get("ALGO_SRC_VISIBILITY", "closed")
        self["license"] = os.environ.get("ALGO_LICENSE", "apl")
        self["network_access"] = os.environ.get("ALGO_NETWORK_ACCESS", "full")
        self["pipeline_enabled"] = os.environ.get("ALGO_PIPELINE", True)

class Progress(remote.RemoteProgress):
    def line_dropped(self, line):
        print(line)

    def update(self, op_code, cur_count, max_count=None, message=""):
        print(self._cur_line)


def run_local(name, model_uri, flavor=None, config=None):
    logger.info("Use `mlflow models serve` to run this model locally")


def target_help():
    return "Deploy MLflow models to Algorithmia"



import json

import mlflow
from Algorithmia.errors import AlgorithmException
from mlflow import pyfunc
from mlflow.pyfunc import scoring_server


try:
    from . import algorithmia_utils
except ImportError:
    import algorithmia_utils


class MLflowWrapper(object):
    def __init__(self, model_fpath=None):
        self.model = None

        if model_fpath:
            self.load_model(model_fpath)

    def load_model(self, model_fpath):
        if model_fpath.startswith("data://"):
            model_fpath = algorithmia_utils.get_file(model_fpath)

        self.model = pyfunc.load_model(model_fpath)

    def predict(self, input):
        if isinstance(input, dict):
            input = json.dumps(input)

        if isinstance(input, str):
            df = scoring_server.parse_json_input(input)
        else:
            raise AlgorithmException("Input should be str or json")

        return self.model.predict(df)



# Test for reading and parsing conda.yaml

from mlflow_algorithmia.conda_env import Environment


def test_conda_yaml_1():
    env = Environment.from_yamlstr(
        """
dependencies:
- python=3.6.12
- scikit-learn=0.23.2
- pip
- pip:
  - mlflow
  - cloudpickle==1.6.0
name: mlflow-env
"""
    )
    assert env.name == "mlflow-env"
    assert env.channels == []
    assert env.list_deps() == ["scikit-learn==0.23.2", "mlflow", "cloudpickle==1.6.0"]


def test_conda_yaml_2():
    env = Environment.from_yamlstr(
        """
channels:
- defaults
- conda-forge
dependencies:
- python=3.6.12
- scikit-learn=0.23.2
- tensorflow>=2.0
- pytorch>1.0,<2.0
- boto
- pip
- pip:
  - mlflow
  - dask>1.0
  - cloudpickle==1.6.0
name: mlflow-complex-env
"""
    )
    assert env.name == "mlflow-complex-env"
    assert env.channels == ["defaults", "conda-forge"]
    assert env.list_deps() == [
        "scikit-learn==0.23.2",
        "tensorflow>==2.0",
        "pytorch<2.0,>1.0",
        "boto",
        "mlflow",
        "dask>1.0",
        "cloudpickle==1.6.0",
    ]



import os

from setuptools import find_packages, setup


setup_dir = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    filepath = os.path.join(setup_dir, filename)
    with open(filepath) as file:
        return file.read()


setup(
    name="mlflow-algorithmia",
    use_scm_version=True,
    packages=find_packages(),
    # package_dir={"": "src"},
    zip_safe=False,
    include_package_data=True,
    # package_data={"mlflow_algorithmia": ["includes/*"]},
    # data_files=data_files,
    # cmdclass={"install": InstallCmd},
    entry_points={
        "mlflow.deployments": "algorithmia=mlflow_algorithmia.deployment",
    },
    options={"bdist_wheel": {"universal": "1"}},
    python_requires=">=3.6",
    setup_requires=["setuptools_scm"],
    install_requires=read_file("requirements-package.txt").splitlines(),
    extras_require={
        "test": ["pytest"],
        "dev": read_file("requirements.txt").splitlines(),
    },
    description="",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    license="Apache License, Version 2.0",
    maintainer="Algorithmia",
    maintainer_email="drodriguez@algorithmia.io",
    url="https://github.com/algorithmia/mlflow-algorithmia",
    keywords=[],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



model = None
mlflow_bundle = "{{ mlflow_bundle_file }}"


def get_model():
    global model
    if model is None:
        try:
            from .mlflow_wrapper import MLflowWrapper
        except ImportError:
            from mlflow_wrapper import MLflowWrapper

        model = MLflowWrapper()
        model.load_model(mlflow_bundle)
    return model


def apply(input):
    try:
        model = get_model()
        predictions = model.predict(input)
        return predictions.tolist()
    except Exception as ex:
        raise ex



import os
import subprocess

import Algorithmia


client = Algorithmia.client()

api_key = os.environ.get("ALGO_API_KEY", None)
if api_key is not None:
    client = Algorithmia.client(api_key)
else:
    client = Algorithmia.client()


in_algorithmia = True if os.environ.get("ALGORITHMIA_API", False) else False


def extract_tar_gz(file, output_dir="./models"):
    """
    Extract a .tar.gz
    Parameters
    ----------
        output_dir (default="./models"): Where to extract the .tar.gz
    Returns
    ------
        output_dir: full path to the output directory where files where extracted
    """
    os.makedirs(output_dir, exist_ok=True)

    try:
        output = subprocess.check_output(
            "tar -C {output} -xzf {targz}".format(output=output_dir, targz=file),
            stderr=subprocess.STDOUT,
            shell=True,
        ).decode()
    except subprocess.CalledProcessError as ex:
        output = ex.output.decode()
        raise Exception("Could not extract file: %s" % output)

    return os.path.realpath(os.path.join(output_dir))


def get_file(remote_fpath):
    """
    Download a file hosted on Algorithmia Hosted Data
    If the file ends with .tar.gz it will untar the file.
    It's recommended that the tar file contain a single files compressed like:
        tar -czvf model.format.tar.gz model.format
    Returns the local file path of the downloaded file
    """
    basename = os.path.basename(remote_fpath)

    if remote_fpath.startswith("data://"):
        # Download from Algoritmia hosted data
        local_fpath = client.file(remote_fpath).getFile().name

        if basename.endswith(".tar.gz"):
            output_dir = extract_tar_gz(local_fpath)
            no_ext = basename[: -len(".tar.gz")]
            local_fpath = os.path.join(output_dir, no_ext)

        return local_fpath

    return remote_fpath


def exists(username, collection, fname=None, connector="data"):
    if fname is None:
        path = f"{connector}://{username}/{collection}"
        obj = client.dir(path)
        return obj.exists()
    else:
        path = f"{connector}://{username}/{collection}/{fname}"
        obj = client.file(path)
        return obj.exists()


def upload_file(
    local_filename,
    username,
    collection,
    fname,
    connector="data",
):
    dir_exists = exists(username=username, collection=collection, connector=connector)
    if dir_exists is False:
        dir_path = f"{connector}://{username}/{collection}/"
        new_dir = client.dir(dir_path)
        new_dir.create()

    remote_file = f"{connector}://{username}/{collection}/{fname}"
    client.file(remote_file).putFile(local_filename)
    return remote_file



import pytest
from mlflow.exceptions import MlflowException

from mlflow_algorithmia.deployment import Settings


@pytest.fixture
def mock_env_set(monkeypatch):
    monkeypatch.setenv("ALGORITHMIA_API_KEY", "algo-key")
    monkeypatch.setenv("ALGORITHMIA_USERNAME", "test_user")


@pytest.fixture
def mock_env_missing(monkeypatch):
    monkeypatch.delenv("ALGORITHMIA_API_KEY", raising=False)
    monkeypatch.delenv("ALGORITHMIA_USERNAME", raising=False)


def test_missing_api_key(mock_env_missing):
    with pytest.raises(MlflowException):
        s = Settings()


def test_missing_username(mock_env_set, monkeypatch):
    monkeypatch.delenv("ALGORITHMIA_USERNAME", raising=False)
    with pytest.raises(MlflowException):
        s = Settings()


def test_git_endpoint(mock_env_set):
    s = Settings()
    assert s["git_endpoint"] == "git.algorithmia.com"


def test_git_endpoint_test(mock_env_set, monkeypatch):
    monkeypatch.setenv("ALGORITHMIA_API", "https://api.test.algorithmia.com")
    s = Settings()
    assert s["git_endpoint"] == "git.test.algorithmia.com"


def test_git_endpoint_other(mock_env_set, monkeypatch):
    monkeypatch.setenv("ALGORITHMIA_API", "https://api.algorithmia.mycorp.net")
    s = Settings()
    assert s["git_endpoint"] == "git.algorithmia.mycorp.net"
